#!/usr/bin/perl
#
# Kakusan
# This is a script to select nucleotide substitution model for multi-
# partitioned data set. Official web site of this script is
# http://www.fifthdimension.jp/products/kakusan/ .
# To know script details, see above URL.
# 
# Copyright (C) 2006-2007  Akifumi S. Tanabe
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

use strict;

$| = 1;

my $readseq = 'readseq';
my $jreadseq = 'java -jar readseq.jar';
my $paupcmd = 'paup';

my @allmodel = ('JC69', 'F81', 'K80', 'HKY85', 'TN93ef', 'TN93', 'K81', 'K81uf', 'TIMef', 'TIM', 'TVMef', 'TVM', 'SYM', 'GTR');
my @allheterogeneity = ('homogeneous', 'gamma', 'invariant', 'gammainvariant', 'arbitrary');

my @modelformrbayes = ('JC69', 'F81', 'K80', 'HKY85', 'SYM', 'GTR');
my @heterogeneityformrbayes = ('homogeneous', 'gamma', 'invariant', 'gammainvariant', 'arbitrary');

my @modelforpaup = ('JC69', 'F81', 'K80', 'HKY85', 'TN93ef', 'TN93', 'K81', 'K81uf', 'TIMef', 'TIM', 'TVMef', 'TVM', 'SYM', 'GTR');
my @heterogeneityforpaup = ('homogeneous', 'gamma', 'invariant', 'gammainvariant', 'arbitrary');

my $fullpathname;
my $treefile;
my $inputfile;
my $chisq = 1;
my $siglevel = 0.05;
my $basefreq = 'Estimate';
my $ngammacat = 4;
my $numthreads = 1;
my $difftreeeachpart;
my $algorithm = 'NJJC';
my $dividecodon = 1;
my $calcwhole = 1;
my $ntax;
my $nchar;
my $maindir;
my @taxa;
my @partname;
my @subpart;
my @allpart;
my %part2site;
my %subpart;
my %models;
my %heterogeneities;
my %model;
my %heterogeneity;
my %nparam;
my %samplesize;
my %LnL;
my %criteria;
my %weight;
my %output;
my %notarbitrary;

# Main routine

my $buildno = '1.0.2007.02.27';

print << "_END";
Kakusan $buildno
======================

This is a script to select nucleotide substitution model for multi-
partitioned data set. Official web site of this script is
http://www.fifthdimension.jp/products/kakusan/ .
To know script details, see above URL.

Copyright (C) 2006-2007  Akifumi S. Tanabe

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

_END

# parse command line options
&parseOptions();

# split protein coding sequences to each codon position
&splitCodingSequences();

# check required and obstructive files
&checkFiles();

# convert sequences to non-interleaved NEXUS format
&convertSequences();

# split sequences to specified partitions
&splitSequences();

# set candidate models
&setCandidateModels();

# run pre-selection chi-square test
if ($chisq) {
	&runChiSquareTest();
}

# set the number of model parameters
&setModelParameters();

# make starting tree
&makeStartingTree();

# set sample sizes
&setSampleSizes();

# analyze partitions
&analyzeEachPartition();

# calculate model selection criteria
&calculateCriteria();

# output files
&outputFiles();

# exit
exit(0);

sub parseOptions {
	print("Parsing command line options...\n");
	my %site2part;
	foreach (@ARGV) {
		if (/^-+(?:help|h)/i) {
			&helpMessage();
		}
		# -c option
		elsif (/^-+(?:chisq|c)=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$chisq = 1;
			} elsif ($1 =~ /^Disable$/i) {
				$chisq = 0;
			} else {
				&errorMessage('Chi-Square test option is not valid.');
			}
		}
		# -s option
		elsif (/^-+(?:siglevel|s)=(.+)/i) {
			if ($1 =~ /^(0\.?\d+)$/) {
				$siglevel = $1;
			} else {
				&errorMessage('The number of significant level is not valid.');
			}
			if ($siglevel == 0 || $siglevel >= 1) {
				&errorMessage('The number of significant level is not valid.');
			}
		}
		# -d option
		elsif (/^-+(?:dividecodon|d)=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$dividecodon = 1;
			} elsif ($1 =~ /^Disable$/i) {
				$dividecodon = 0;
			} else {
				&errorMessage('Codon position dividing option is not valid.');
			}
		}
		# -w option
		elsif (/^-+(?:calcwhole|w)=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$calcwhole = 1;
			} elsif ($1 =~ /^Disable$/i) {
				$calcwhole = 0;
			} else {
				&errorMessage('The "calcwhole" option is not valid.');
			}
		}
		# -f option
		elsif (/^-+(?:basefreq|f)=(.+)/i) {
			if ($1 =~ /^Empirical$/i) {
				$basefreq = 'Empirical';
			} elsif ($1 =~ /^Estimate$/i) {
				$basefreq = 'Estimate';
			} else {
				&errorMessage('BaseFreq option is not valid.');
			}
		}
		# -p option
		elsif (/^-+(?:partition|p)=(.+)/i) {
			foreach my $partition (split(/,/, $1)) {
				if ($partition =~ s/^([^:]+)://) {
					my $partname = $1;
					my @range;
					if ($partname eq 'whole') {
						&errorMessage('Partition name "whole" is reserved by Kakusan.');
					}
					elsif ($partname =~ /(_P[1-3])/i) {
						&errorMessage("Partition name \"-$1\" is reserved by Kakusan.");
					}
					unless ($part2site{$partname}) {
						push(@partname, $partname);
					} else {
						&errorMessage("\"$partname\" is doubly specified.");
					}
					# case of normal partition
					if ($partition =~ /^(\d+)-(\d+)$/) {
						@range = &range2list($1,$2,1);
					}
					else {
						&errorMessage('Partition specification is not valid.');
					}
					# Set site numbers to partition name
					foreach my $site (@range) {
						unless ($site2part{$site}) {
							$site2part{$site} = $partname;
						} else {
							&errorMessage("Site $site is overlapping between \"$site2part{$site}\" and \"$partname\".");
						}
						push(@{$part2site{$partname}}, $site);
					}
				} else {
					&errorMessage('Partition name is not specified.');
				}
			}
		}
		# -g option
		elsif (/^-+(?:ngammacat|g)=(.+)/i) {
			if ($1 =~ /^(\d+)$/) {
				$ngammacat = $1;
			} else {
				&errorMessage('The number of rate categories is not an integer.');
			}
			if ($ngammacat < 2) {
				&errorMessage('The number of rate categories is not valid.');
			}
		}
		# -a option
		elsif (/^-+(?:algorithm|a)=(.+)/i) {
			if ($1 =~ /^NJJC$/i) {
				$algorithm = 'NJJC';
			} elsif ($1 =~ /^NJLogDet$/i) {
				$algorithm = 'NJLogDet';
			} elsif ($1 =~ /^SARandom$/i) {
				$algorithm = 'SARandom';
			} elsif ($1 =~ /^SAClosest$/i) {
				$algorithm = 'SAClosest';
			} elsif ($1 =~ /^SAFurthest$/i) {
				$algorithm = 'SAFurthest';
			} else {
				&errorMessage('Algorithm option of tree generation is not valid.');
			}
		}
		# -t option
		elsif (/^-+(?:tree|t)=(.+)/i) {
			if ($1 =~ /^Different$/i) {
				$difftreeeachpart = 1;
			} elsif ($1 =~ /^Common$/i) {
				next;
			} else {
				&errorMessage('Tree option for parameter optimization is not valid.');
			}
		}
		# -o option
		elsif (/^-+(?:output|o)=(.+)/i) {
			foreach (split(/,/, $1)) {
				if (/^MrBayes$/i) {
					foreach (@modelformrbayes) {
						$model{$_} = 1;
					}
					foreach (@heterogeneityformrbayes) {
						$heterogeneity{$_} = 1;
					}
					$output{'mrbayes'} = 1;
				} elsif (/^PAUP$/i) {
					foreach (@modelforpaup) {
						$model{$_} = 1;
					}
					foreach (@heterogeneityforpaup) {
						$heterogeneity{$_} = 1;
					}
					$output{'paup'} = 1;
				} elsif (/^all$/i) {
					foreach (@allmodel) {
						$model{$_} = 1;
					}
					foreach (@allheterogeneity) {
						$heterogeneity{$_} = 1;
					}
					$output{'all'} = 1;
				} else {
					&errorMessage('Specified output option is not valid.');
				}
			}
		}
		# -n option
		elsif (/^-+(?:numthreads|n)=(.+)/i) {
			if ($1 =~ /^(\d+)$/) {
				$numthreads = $1;
			} else {
				&errorMessage('The number of threads is not an integer.');
			}
			if ($numthreads == 0) {
				&errorMessage('The number of threads is not valid.');
			}
		}
		# --treefile option
		elsif (/^-+treefile=(.+)/i) {
			if (-e "$1") {
				$treefile = $1;
			} else {
				&errorMessage("Specified tree file \"$1\" does not exist.");
			}
		}
		# input file
		else {
			if (-e "$_") {
				$fullpathname = $_;
			} else {
				&errorMessage("Specified input file \"$_\" does not exist.");
			}
		}
	}
	# Is the input file specified?
	unless ($fullpathname) {
		&errorMessage('The input file is not specified.');
	} else {
		my @temparray = $fullpathname =~ /([^\/\\]+)$/;
		$inputfile = shift(@temparray);
		$maindir = $inputfile . '.kakusan';
	}
	# 
	if ($calcwhole == 0 && @partname < 2) {
		&errorMessage('The "calcwhole" option is not valid.');
	}
	# Are there unused sites?
	for (my $i = 1; $i <= keys(%site2part); $i ++) {
		unless ($site2part{$i}) {
			&errorMessage('Partition specification is not valid. There are unused sites.');
		}
	}
	# If output option is not specified...
	unless (%model) {
		foreach (@allmodel) {
			$model{$_} = 1;
		}
		foreach (@allheterogeneity) {
			$heterogeneity{$_} = 1;
		}
		$output{'all'} = 1;
	}
	print("done.\n\n");
	# Output to screen
	my $lctime = localtime();
	print << "_END";
Time: $lctime
Input File Name: $inputfile
Chi-Square Test: $chisq
Significant Level: $siglevel
Calc whole: $calcwhole
Divide codon: $dividecodon
Base Frequencies: $basefreq
Rate Categories: $ngammacat
Algorithm: $algorithm
Different Tree on each Partition: $difftreeeachpart
Threads: $numthreads
Partitions: @partname
_END
	foreach my $partname (@partname) {
		print("$partname: @{$part2site{$partname}}\n");
	}
	print("Matrix:");
	foreach my $model (keys(%model)) {
		print(" $model");
	}
	print("\n");
	print("Heterogeneity:");
	foreach my $heterogeneity (keys(%heterogeneity)) {
		print(" $heterogeneity");
	}
	print("\n");
	print("\n");
}

sub splitCodingSequences {
	foreach my $partname (@partname) {
		if ($partname =~ /_P$/i) {
			for (my $i = 0; $i < 3; $i ++) {
				push(@subpart, $partname . ($i + 1));
				my @range = &range2list($part2site{$partname}[$i], $part2site{$partname}[-1], 3);
				foreach my $site (@range) {
					push(@{$subpart{$partname . ($i + 1)}}, $site);
				}
			}
		}
	}
}

sub range2list {
	# Input: Site of range start, Site of range end, Skip number of sites
	# Output: List of sites which belong to the range
	my ($start, $end, $skip) = @_;
	my @num;
	if ($start != 0 && $skip != 0 && $start <= $end && $skip <= $end - $start) {
		for (my $i = $start; $i <= $end; $i += $skip) {
			push(@num, $i);
		}
	} else {
		&errorMessage('Partition specification is not valid.');
	}
	return(@num);
}

sub checkFiles {
	print("Checking commands and files...\n");
	# Check required comands
	if (system("$jreadseq -h")) {
		if (system("$readseq -h")) {
			if (system("./$readseq -h")) {
				&errorMessage('Cannot use readseq. Please install readseq.');
			}
		}
	} else {
		$readseq = $jreadseq;
	}
	unless (system("$paupcmd -u -n")) {
		unless (system("./$paupcmd -u -n")) {
			&errorMessage('Cannot run $paupcmd. Please check $paupcmd.');
		}
	}
	# Check working directory
	my $tempno;
	while (-e "$maindir") {
		$tempno ++;
		unless (-e "bkup$tempno.$maindir") {
			print("$maindir already exists. Can I rename it to \"bkup$tempno.$maindir\"?\n");
			my $response;
			chomp($response = <STDIN>);
			if ($response =~ /^yes$/i || $response =~ /^y$/i) {
				unless (rename($maindir, "bkup$tempno.$maindir")) {
					&errorMessage('Cannot change the directory name.');
				} else {
					print("OK. Changed to \"bkup$tempno.$maindir\".\n");
				}
			} else {
				&errorMessage('The script aborted.');
			}
		}
	}
	# Make working directory
	unless (mkdir($maindir)) {
		&errorMessage("Cannot make \"$maindir\" directory.");
	}
	unless (chdir($maindir)) {
		&errorMessage('Cannot change working directory.');
	}
	unless (mkdir('Logs')) {
		&errorMessage("Cannot make \"$maindir/Logs\" directory.");
	}
	unless (chdir('..')) {
		&errorMessage('Cannot change working directory.');
	}
	print("done.\n\n");
}

sub convertSequences {
	print("Converting sequence file...\n");
	# Preprocess input file
	unless (open(INPUT, "< $fullpathname")) {
		&errorMessage("Cannot open \"$fullpathname\".");
	}
	my @seqdata;
	while (<INPUT>) {
		push(@seqdata, split(/(?:\r\n|\r|\n)/));
	}
	close(INPUT);
	unless (open(OUTPUT, "> $maindir/$inputfile")) {
		&errorMessage("Cannot make \"$maindir/$inputfile\".");
	}
	my $nexusornot;
	foreach (@seqdata) {
		if (/#NEXUS/i) {
			print(OUTPUT "$_\n");
			$nexusornot = 1;
		}
		if ($nexusornot == 1 && (/Begin\s+Data\s*;/i || /Begin\s+Characters\s*;/i)) {
			$nexusornot = 2;
		}
		if ($nexusornot == 2 && /NTax=(\d+)/i) {
			$ntax = $1;
		}
		if ($nexusornot == 2 && /NChar=(\d+)/i) {
			$nchar = $1;
		}
		if ($nexusornot == 2 && /End\s*;/i) {
			print(OUTPUT "$_\n");
			$nexusornot = 1;
		}
		if ($nexusornot != 1) {
			print(OUTPUT "$_\n");
		}
	}
	if ($nexusornot && $ntax && $nchar) {
		print(OUTPUT "Begin PAUP;\n\tExport File=$maindir/data_whole.nex Format=NEXUS;\n\tQuit;\nEnd;\n");
	}
	close(OUTPUT);
	if ($nexusornot && $ntax && $nchar) {
		if (system("$paupcmd $maindir/$inputfile > $maindir/Logs/convert.txt")) {
			&errorMessage("Cannot run $paupcmd. Please check $paupcmd.");
		}
	} else {
		# Convert to FASTA by ReadSeq
		if (system("$readseq -a -C -f8 -o=$maindir/input.fas $maindir/$inputfile")) {
			&errorMessage('Cannot convert sequence file.');
		}
		# Convert ReadSeq output FASTA to NEXUS
		unless (open(FASTA, "< $maindir/input.fas")) {
			&errorMessage("Cannot open \"$maindir/input.fas\".");
		}
		while (<FASTA>) {
			if (/^>\w+.*?,? (\d+) .+\n$/) {
				$ntax ++;
				unless ($nchar) {
					$nchar = $1;
				} else {
					if ($nchar != $1) {
						&errorMessage('Sequence lengths are not equal.');
					}
				}
			}
		}
		close(FASTA);
		if ($ntax < 3) {
			&errorMessage('The number of taxon is too small or file parsing error.');
		}
		unless (open(NEXUS, "> $maindir/data_whole.nex")) {
			&errorMessage("Cannot make \"$maindir/data_whole.nex\".");
		}
		print(NEXUS '#NEXUS' . "\n\nBegin Data;\n\tDimensions NTax=$ntax NChar=$nchar;\n\tFormat DataType=DNA Gap=- Missing=?;\n\tMatrix");
		unless (open(FASTA, "< $maindir/input.fas")) {
			&errorMessage("Cannot open \"$maindir/input.fas\".");
		}
		while (<FASTA>) {
			s/^>(\w+).*?,? \d+ .+\r?\n?$/\n$1 /;
			s/^([^>][^\r\n]*?)\r?\n?$/$1/;
			print(NEXUS $_);
		}
		close(FASTA);
		print(NEXUS "\n\t;\nEnd;\n");
		close(NEXUS);
	}
	print("done.\n\n");
}

sub splitSequences {
	print("Splitting sequences...\n");
	unless (chdir($maindir)) {
		&errorMessage('Cannot change working directory.');
	}
	# Read sequence file and output to partitioned file
	unless (open(NEXUS, "< data_whole.nex")) {
		&errorMessage("Cannot open \"$maindir/data_whole.nex\".");
	}
	my %seqs;
	my $line = 0;
	while (<NEXUS>) {
		$line ++;
		if ($line < 7) {
			next;
		} elsif ($line >= 7 && $line < 7 + $ntax) {
			if (/^(\S+)\s+(\S+)\r?\n?$/) {
				push(@taxa, $1);
				for (my $i = 0; $i < $nchar; $i ++) {
					push(@{$seqs{$1}}, uc(substr($2, $i, 1)));
				}
			} else {
				&errorMessage("\"$maindir/data_whole.nex\" is not valid.");
			}
		} elsif ($line == 7 + $ntax) {
			last;
		}
	}
	close(NEXUS);
	push(@allpart, 'whole');
	if (@partname > 1) {
		foreach my $partname (@partname) {
			push(@allpart, $partname);
			unless (open(OUT, "> data_$partname.nex")) {
				&errorMessage("Cannot make \"$maindir/data_$partname.nex\".");
			}
			print(OUT '#NEXUS' . "\n\nBegin Data;\n\tDimensions NTax=$ntax NChar=" . scalar(@{$part2site{$partname}}) . ";\n\tFormat DataType=DNA Gap=- Missing=?;\n\tMatrix\n");
			foreach my $taxa (@taxa) {
				print(OUT $taxa . ' ');
				foreach my $site (@{$part2site{$partname}}) {
					print(OUT $seqs{$taxa}[$site - 1]);
				}
				print(OUT "\n");
			}
			print(OUT "\t;\nEnd;\n");
			close(OUT);
		}
	}
	if ($dividecodon) {
		foreach my $partname (@subpart) {
			push(@allpart, $partname);
			unless (open(OUT, "> data_$partname.nex")) {
				&errorMessage("Cannot make \"$maindir/data_$partname.nex\".");
			}
			print(OUT '#NEXUS' . "\n\nBegin Data;\n\tDimensions NTax=$ntax NChar=" . scalar(@{$subpart{$partname}}) . ";\n\tFormat DataType=DNA Gap=- Missing=?;\n\tMatrix\n");
			foreach my $taxa (@taxa) {
				print(OUT $taxa . ' ');
				foreach my $site (@{$subpart{$partname}}) {
					print(OUT $seqs{$taxa}[$site - 1]);
				}
				print(OUT "\n");
			}
			print(OUT "\t;\nEnd;\n");
			close(OUT);
		}
	}
	unless (chdir('..')) {
		&errorMessage('Cannot change working directory.');
	}
	print("done.\n\n");
}

sub setCandidateModels {
	# Candidate models of each partition
	foreach my $partname (@allpart) {
		foreach my $model (keys(%model)) {
			$models{$partname}{$model} = 1;
		}
		foreach my $heterogeneity (keys(%heterogeneity)) {
			if ($heterogeneity ne 'arbitrary' || ($partname eq 'whole' || $partname =~ /_P$/i)) {
				$heterogeneities{$partname}{$heterogeneity} = 1;
			}
		}
	}
}

sub runChiSquareTest {
	print("Testing homogeneity of base frequencies across taxa...\n");
	unless (chdir($maindir)) {
		&errorMessage('Cannot change working directory.');
	}
	foreach my $partname (@allpart) {
		unless (open(CMD, "> chisq_$partname.nex")) {
			&errorMessage("Cannot make \"$maindir/chisq_$partname.nex\".");
		}
		print(CMD '#NEXUS' . "\n\nBegin PAUP;\n\tSet ErrorBeep=No QueryBeep=No KeyBeep=No NotifyBeep=No ErrorStop=No WarnReset=No WarnTree=No WarnTSave=No WarnBlkName=No WarnRoot=No WarnRedef=No;\n\tExecute data_$partname.nex;\n\tBaseFreqs;\n\tQuit;\nEnd;\n");
		close(CMD);
		if (system("$paupcmd chisq_$partname.nex > Logs/chisq_$partname.txt")) {
			&errorMessage("Cannot run $paupcmd. Please check $paupcmd.");
		}
		unless (rename("chisq_$partname.nex", "Logs/chisq_$partname.nex")) {
			&errorMessage("Cannot move \"$maindir/chisq_$partname.nex\".");
		}
		unless (open(LOG, "< Logs/chisq_$partname.txt")) {
			&errorMessage("Cannot read \"$maindir/Logs/chisq_$partname.txt\".");
		}
		my $p_value;
		while (<LOG>) {
			if (/^Chi-square = \d+\.\d+ \(df=\d+\), P = ([01]\.\d+)/) {
				$p_value = $1;
				last;
			}
		}
		close(LOG);
		if ($p_value <= $siglevel) {
			delete($models{$partname}{'F81'});
			delete($models{$partname}{'HKY85'});
			delete($models{$partname}{'TN93'});
			delete($models{$partname}{'K81uf'});
			delete($models{$partname}{'TIM'});
			delete($models{$partname}{'TVM'});
			delete($models{$partname}{'GTR'});
		}
	}
	unless (chdir('..')) {
		&errorMessage('Cannot change working directory.');
	}
	print("done.\n\n");
}

sub setModelParameters {
	print("Counting model parameters...\n");
	foreach my $partname (@allpart) {
		foreach my $model (keys(%{$models{$partname}})) {
			foreach my $heterogeneity (keys(%{$heterogeneities{$partname}})) {
				# the number of branch lengths
				$nparam{$partname}{$model . '_' . $heterogeneity} = ($ntax * 2) - 3;
				# the number of parameters of nucleotide substitution rate matrix
				if ($model eq 'JC69') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 0;
				} elsif ($model eq 'F81') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 3;
				} elsif ($model eq 'K80') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 1;
				} elsif ($model eq 'HKY85') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 4;
				} elsif ($model eq 'TN93ef') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 2;
				} elsif ($model eq 'TN93') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 5;
				} elsif ($model eq 'K81') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 2;
				} elsif ($model eq 'K81uf') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 5;
				} elsif ($model eq 'TIMef') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 3;
				} elsif ($model eq 'TIM') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 6;
				} elsif ($model eq 'TVMef') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 4;
				} elsif ($model eq 'TVM') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 7;
				} elsif ($model eq 'SYM') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 5;
				} elsif ($model eq 'GTR') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 8;
				}
				# the number of parameters of site rate heterogeneity
				if ($heterogeneity eq 'homogeneous') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 0;
				} elsif ($heterogeneity eq 'gamma') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 1;
				} elsif ($heterogeneity eq 'invariant') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 1;
				} elsif ($heterogeneity eq 'gammainvariant') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 2;
				} elsif ($heterogeneity eq 'arbitrary' && $partname =~ /_P$/i) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 2;
				} elsif ($heterogeneity eq 'arbitrary' && $partname eq 'whole') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += (@partname - (@subpart / 3) + @subpart) - 1;
				} else {
					&errorMessage('Model setting error.');
				}
			}
		}
	}
#	foreach (keys(%nparam)) {
#		print("$_ : $nparam{$_}\n");
#	}
	print("done.\n\n");
}

sub makeStartingTree {
	print("Generating tree topology for parameter optimization...\n");
	unless (chdir($maindir)) {
		&errorMessage('Cannot change working directory.');
	}
	if ($difftreeeachpart && @partname > 1) {
		my $child = 0;
		$? = 0;
		foreach my $partname ('whole',@partname) {
			if (my $pid = fork()) {
				$child ++;
				if ($child == $numthreads) {
					if (wait == -1) {
						$child = 0;
					} else {
						$child --;
					}
				}
				if ($?) {
					&errorMessage('Tree building error.');
				}
				next;
			} else {
				&starttreecmd($partname);
				exit(0);
			}
		}
		while (wait != -1) {
			if ($?) {
				&errorMessage('Tree building error.');
			}
		}
	} else {
		&starttreecmd('whole');
	}
	unless (chdir('..')) {
		&errorMessage('Cannot change working directory.');
	}
	print("done.\n\n");
}

sub starttreecmd {
	my $partname = shift;
	unless (open(CMD, "> maketree_$partname.nex")) {
		&errorMessage("Cannot make \"$maindir/maketree_$partname.nex\".");
	}
	print(CMD '#NEXUS' . "\n\nBegin PAUP;\n\tSet ErrorBeep=No QueryBeep=No KeyBeep=No NotifyBeep=No ErrorStop=No WarnReset=No WarnTree=No WarnTSave=No WarnBlkName=No WarnRoot=No WarnRedef=No;\n\tExecute data_$partname.nex;\n");
	if ($algorithm eq 'NJJC') {
		print(CMD "\tDSet Distance=JC BaseFreq=Equal Rates=Equal PInvar=0;\n\tNJ ShowTree=No BreakTies=Random TreeFile=starttree_$partname.tre Replace=Yes;\n");
	} elsif ($algorithm eq 'NJLogDet') {
		print(CMD "\tDSet Distance=LogDet BaseFreq=Equal Rates=Equal PInvar=0;\n\tNJ ShowTree=No BreakTies=Random TreeFile=starttree_$partname.tre Replace=Yes;\n");
	} elsif ($algorithm eq 'SARandom') {
		print(CMD "\tHSearch Start=Stepwise AddSeq=Random NReps=1 Swap=None MulTrees=No;\n\tSaveTrees Format=AltNexus File=starttree_$partname.tre Replace=Yes;\n");
	} elsif ($algorithm eq 'SAClosest') {
		print(CMD "\tHSearch Start=Stepwise AddSeq=Closest Swap=None MulTrees=No;\n\tSaveTrees Format=AltNexus File=starttree_$partname.tre Replace=Yes;\n");
	} elsif ($algorithm eq 'SAFurthest') {
		print(CMD "\tHSearch Start=Stepwise AddSeq=Furthest Swap=None MulTrees=No;\n\tSaveTrees Format=AltNexus File=starttree_$partname.tre Replace=Yes;\n");
	} else {
		&errorMessage('Cannot recognize algorithm.');
	}
	print(CMD "\tQuit WarnTSave=No;\nEnd;\n");
	close(CMD);
	if (system("$paupcmd maketree_$partname.nex > Logs/maketree_$partname.txt")) {
		&errorMessage("Cannot run $paupcmd. Please check $paupcmd.");
	}
	unless (rename("maketree_$partname.nex", "Logs/maketree_$partname.nex")) {
		&errorMessage("Cannot move \"$maindir/maketree_$partname.nex\".");
	}
}

sub setSampleSizes {
	print("Counting sample sizes...\n");
	unless (chdir($maindir)) {
		&errorMessage('Cannot change working directory.');
	}
	# sample size 0 : pscore
	# sample size 1 : sum of (the number of character states - 1)
	# sample size 2 : sum of the number of character states
	# sample size 3 : the number of sites = alignment length
	# sample size 4 : the number of variable characters
	# sample size 5 : (sample size 3) x the number of OTUs
	my $child = 0;
	$? = 0;
	foreach my $partname (@allpart) {
		if (my $pid = fork()) {
			$child ++;
			if ($child == $numthreads) {
				if (wait == -1) {
					$child = 0;
				} else {
					$child --;
				}
			}
			if ($?) {
				&errorMessage('Sample size counting error.');
			}
			next;
		} else {
			&calcpscore($partname);
			exit(0);
		}
	}
	while (wait != -1) {
		if ($?) {
			&errorMessage('Sample size counting error.');
		}
	}
	foreach my $partname (@allpart) {
		unless (open(SCORE, "< pscore_$partname.txt")) {
			&errorMessage("Cannot open \"$maindir/pscore_$partname.txt\".");
		}
		while (<SCORE>) {
			if (/^1\t(\d+)/) {
				$samplesize{$partname}[0] = $1;
				last;
			}
		}
		close(SCORE);
		unless (unlink("pscore_$partname.txt")) {
			&errorMessage("Cannot delete \"$maindir/pscore_$partname.txt\".");
		}
	}
	unless (open(NEXUS, "< data_whole.nex")) {
		&errorMessage("Cannot open \"$maindir/data_whole.nex\".");
	}
	my %tempseqs;
	my $line = 0;
	while (<NEXUS>) {
		$line ++;
		if ($line < 7) {
			next;
		} else {
			if (/^(\S+)\s+(\S+)\r?\n?$/) {
				for (my $i = 0; $i < length($2); $i ++) {
					push(@{$tempseqs{$1}}, uc(substr($2, $i, 1)));
				}
			} elsif (/^\t;\r?\n?$/) {
				last;
			} else {
				&errorMessage("\"$maindir/data_whole.nex\" is not valid.");
			}
		}
	}
	close(NEXUS);
	my @numstate;
	for (my $i = 0; $i < $nchar; $i ++) {
		my %state;
		foreach my $taxa (@taxa) {
			if ($tempseqs{$taxa}[$i] =~ /^[ACGTMRWSYKVHDBN]$/i) {
				$state{$tempseqs{$taxa}[$i]} = 1;
			} elsif ($tempseqs{$taxa}[$i] =~ /^U$/) {
				$state{'T'} = 1;
			} elsif ($tempseqs{$taxa}[$i] =~ /^-$/) {
				$state{'N'} = 1;
			} elsif ($tempseqs{$taxa}[$i] =~ /^\?$/) {
				$state{'N'} = 1;
			}
		}
		if (exists($state{'M'}) && (exists($state{'A'}) || exists($state{'C'}))) {
			delete($state{'M'});
		}
		if (exists($state{'R'}) && (exists($state{'A'}) || exists($state{'G'}))) {
			delete($state{'R'});
		}
		if (exists($state{'W'}) && (exists($state{'A'}) || exists($state{'T'}))) {
			delete($state{'W'});
		}
		if (exists($state{'S'}) && (exists($state{'C'}) || exists($state{'G'}))) {
			delete($state{'S'});
		}
		if (exists($state{'Y'}) && (exists($state{'C'}) || exists($state{'T'}))) {
			delete($state{'Y'});
		}
		if (exists($state{'K'}) && (exists($state{'G'}) || exists($state{'T'}))) {
			delete($state{'K'});
		}
		if (exists($state{'V'}) && (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'M'}) || exists($state{'R'}) || exists($state{'S'}))) {
			delete($state{'V'});
		}
		if (exists($state{'H'}) && (exists($state{'A'}) || exists($state{'C'}) || exists($state{'T'}) || exists($state{'M'}) || exists($state{'W'}) || exists($state{'Y'}))) {
			delete($state{'H'});
		}
		if (exists($state{'D'}) && (exists($state{'A'}) || exists($state{'G'}) || exists($state{'T'}) || exists($state{'R'}) || exists($state{'W'}) || exists($state{'K'}))) {
			delete($state{'D'});
		}
		if (exists($state{'B'}) && (exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'}) || exists($state{'S'}) || exists($state{'Y'}) || exists($state{'K'}))) {
			delete($state{'B'});
		}
		if (exists($state{'V'}) && exists($state{'H'}) && exists($state{'D'}) && exists($state{'B'})) {
			$numstate[$i] = 2;
		} elsif (exists($state{'V'}) && exists($state{'H'}) && exists($state{'D'}) || exists($state{'V'}) && exists($state{'H'}) && exists($state{'B'}) || exists($state{'V'}) && exists($state{'D'}) && exists($state{'B'}) || exists($state{'H'}) && exists($state{'D'}) && exists($state{'B'})) {
			$numstate[$i] = 1;
		} elsif (exists($state{'V'}) && exists($state{'H'})) {
			if (exists($state{'K'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'V'}) && exists($state{'D'})) {
			if (exists($state{'Y'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'V'}) && exists($state{'B'})) {
			if (exists($state{'W'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'H'}) && exists($state{'D'})) {
			if (exists($state{'S'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'H'}) && exists($state{'B'})) {
			if (exists($state{'R'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'D'}) && exists($state{'B'})) {
			if (exists($state{'M'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'V'})) {
			if (exists($state{'T'}) || exists($state{'W'}) && exists($state{'Y'}) || exists($state{'W'}) && exists($state{'K'}) || exists($state{'Y'}) && exists($state{'K'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'H'})) {
			if (exists($state{'G'}) || exists($state{'R'}) && exists($state{'S'}) || exists($state{'R'}) && exists($state{'K'}) || exists($state{'S'}) && exists($state{'K'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'D'})) {
			if (exists($state{'C'}) || exists($state{'M'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'Y'}) || exists($state{'S'}) && exists($state{'Y'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'B'})) {
			if (exists($state{'A'}) || exists($state{'M'}) && exists($state{'R'}) || exists($state{'M'}) && exists($state{'W'}) || exists($state{'R'}) && exists($state{'W'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'})) {
			$numstate[$i] = 3;
		} elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'})) {
			$numstate[$i] = 2;
		} elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'})) {
			$numstate[$i] = 2;
		} elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'K'}) || exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'})) {
			if (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'})) {
				$numstate[$i] = 3;
			} else {
				$numstate[$i] = 2;
			}
		} elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) || exists($state{'M'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'W'}) && exists($state{'Y'}) && exists($state{'K'})) {
			$numstate[$i] = 1;
		} elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'W'}) && exists($state{'S'}) && exists($state{'K'})) {
			$numstate[$i] = 2;
		} elsif (exists($state{'M'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'Y'}) || exists($state{'W'}) && exists($state{'S'})) {
			$numstate[$i] = 2;
		} elsif (exists($state{'M'}) && exists($state{'R'}) || exists($state{'M'}) && exists($state{'W'}) || exists($state{'M'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'W'}) || exists($state{'R'}) && exists($state{'S'}) || exists($state{'R'}) && exists($state{'K'}) || exists($state{'W'}) && exists($state{'Y'}) || exists($state{'W'}) && exists($state{'K'}) || exists($state{'S'}) && exists($state{'Y'}) || exists($state{'S'}) && exists($state{'K'}) || exists($state{'Y'}) && exists($state{'K'})) {
			if (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'M'}) || exists($state{'R'}) || exists($state{'W'}) || exists($state{'S'}) || exists($state{'Y'}) || exists($state{'K'})) {
			if (exists($state{'A'}) && exists($state{'C'}) || exists($state{'A'}) && exists($state{'G'}) || exists($state{'A'}) && exists($state{'T'}) || exists($state{'C'}) && exists($state{'G'}) || exists($state{'C'}) && exists($state{'T'}) || exists($state{'G'}) && exists($state{'T'})) {
				$numstate[$i] = 3;
			} elsif (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'})) {
				$numstate[$i] = 2;
			} else {
				$numstate[$i] = 1;
			}
		} elsif (exists($state{'A'}) && exists($state{'C'}) && exists($state{'G'}) && exists($state{'T'})) {
			$numstate[$i] = 4;
		} elsif (exists($state{'A'}) && exists($state{'C'}) && exists($state{'G'}) || exists($state{'A'}) && exists($state{'C'}) && exists($state{'T'}) || exists($state{'A'}) && exists($state{'G'}) && exists($state{'T'}) || exists($state{'C'}) && exists($state{'G'}) && exists($state{'T'})) {
			$numstate[$i] = 3;
		} elsif (exists($state{'A'}) && exists($state{'C'}) || exists($state{'A'}) && exists($state{'G'}) || exists($state{'A'}) && exists($state{'T'}) || exists($state{'C'}) && exists($state{'G'}) || exists($state{'C'}) && exists($state{'T'}) || exists($state{'G'}) && exists($state{'T'})) {
			$numstate[$i] = 2;
		} elsif (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'}) || exists($state{'N'})) {
			$numstate[$i] = 1;
		}
		$samplesize{'whole'}[2] += $numstate[$i];
		unless ($numstate[$i]) {
			&errorMessage("The number of state of $i + 1 th column is unable to count.");
		} elsif ($numstate[$i] > 1) {
			$samplesize{'whole'}[4] ++;
		}
	}
	$samplesize{'whole'}[3] = $nchar;
	$samplesize{'whole'}[1] = $samplesize{'whole'}[2] - $samplesize{'whole'}[3];
	$samplesize{'whole'}[5] = $nchar * $ntax;
	unless (defined($samplesize{'whole'}[4])) {
		$samplesize{'whole'}[4] = 0;
	}
	if (@partname > 1) {
		foreach my $partname (@partname) {
			foreach my $site (@{$part2site{$partname}}) {
				$samplesize{$partname}[2] += $numstate[$site - 1];
				if ($numstate[$site - 1] > 1) {
					$samplesize{$partname}[4] ++;
				}
			}
			$samplesize{$partname}[3] = scalar(@{$part2site{$partname}});
			$samplesize{$partname}[1] = $samplesize{$partname}[2] - $samplesize{$partname}[3];
			$samplesize{$partname}[5] = scalar(@{$part2site{$partname}}) * $ntax;
			unless (defined($samplesize{$partname}[4])) {
				$samplesize{$partname}[4] = 0;
			}
		}
	}
	if ($dividecodon) {
		foreach my $partname (@subpart) {
			foreach my $site (@{$subpart{$partname}}) {
				$samplesize{$partname}[2] += $numstate[$site - 1];
				if ($numstate[$site - 1] > 1) {
					$samplesize{$partname}[4] ++;
				}
			}
			$samplesize{$partname}[3] = scalar(@{$subpart{$partname}});
			$samplesize{$partname}[1] = $samplesize{$partname}[2] - $samplesize{$partname}[3];
			$samplesize{$partname}[5] = scalar(@{$subpart{$partname}}) * $ntax;
			unless (defined($samplesize{$partname}[4])) {
				$samplesize{$partname}[4] = 0;
			}
		}
	}
	foreach my $partname (@allpart) {
#		print("$partname : @{$samplesize{$partname}}\n");
		foreach (0 .. 5) {
			unless (defined($samplesize{$partname}[$_])) {
				&errorMessage('Sample size counting error.');
			}
		}
	}
	unless (chdir('..')) {
		&errorMessage('Cannot change working directory.');
	}
	print("done.\n\n");
}

sub calcpscore {
	my $partname = shift;
	unless (open(CMD, "> pscore_$partname.nex")) {
		&errorMessage("Cannot make \"$maindir/pscore_$partname.nex\".");
	}
	print(CMD '#NEXUS' . "\n\nBegin PAUP;\n\tSet ErrorBeep=No QueryBeep=No KeyBeep=No NotifyBeep=No ErrorStop=No WarnReset=No WarnTree=No WarnTSave=No WarnBlkName=No WarnRoot=No WarnRedef=No;\n\tExecute data_$partname.nex;\n\tGetTrees File=starttree_");
	if ($difftreeeachpart) {
		if ($partname =~ /^(.+_P)[1-3]$/i) {
			print(CMD $1);
		} else {
			print(CMD $partname);
		}
	} else {
		print(CMD 'whole');
	}
	print(CMD ".tre From=1 To=1 Unrooted=Yes;\n");
	print(CMD "\tPScores 1 / ScoreFile=pscore_$partname.txt Replace=Yes;\n");
	print(CMD "\tQuit WarnTSave=No;\nEnd;\n");
	close(CMD);
	if (system("$paupcmd pscore_$partname.nex > Logs/pscore_$partname.txt")) {
		&errorMessage("Cannot run $paupcmd. Please check $paupcmd.");
	}
	unless (rename("pscore_$partname.nex", "Logs/pscore_$partname.nex")) {
		&errorMessage("Cannot move \"$maindir/pscore_$partname.nex\".");
	}
}

sub analyzeEachPartition {
	print("Analyzing each partition under the models...\n");
	unless (chdir($maindir)) {
		&errorMessage('Cannot change working directory.');
	}
	unless (mkdir('Scores')) {
		&errorMessage("Cannot make \"$maindir/Scores\" directory.");
	}
	my $child = 0;
	$? = 0;
	foreach my $partname (@allpart) {
		foreach my $model (keys(%{$models{$partname}})) {
			foreach my $heterogeneity (keys(%{$heterogeneities{$partname}})) {
				if ($partname ne 'whole' || $calcwhole || @partname == 1) {
					if ($heterogeneity ne 'arbitrary' || $partname =~ /_P$/i || $partname eq 'whole' && @partname == 1 && @subpart == 3 || $partname eq 'whole' && @partname > 1) {
						if (my $pid = fork()) {
							$child ++;
							if ($child == $numthreads) {
								if (wait == -1) {
									$child = 0;
								} else {
									$child --;
								}
							}
							if ($?) {
								&errorMessage('Likelihood calculation error.');
							}
							next;
						} else {
							&calcL($partname, $model, $heterogeneity);
							exit(0);
						}
					}
				}
			}
		}
	}
	while (wait != -1) {
		if ($?) {
			&errorMessage('Likelihood calculation error.');
		}
	}
	unless (chdir('..')) {
		&errorMessage('Cannot change working directory.');
	}
	print("done.\n\n");
}

sub calcL {
	my ($partname, $model, $heterogeneity) = @_;
	print("Calculating likelihood under $model" . '_' . "$heterogeneity model on $partname sequences...\n");
	unless (open(CMD, '> ' . $partname . '_' . $model . '_' . $heterogeneity . '.nex')) {
		&errorMessage("Cannot make \"$maindir/$partname" . '_' . $model . '_' . "$heterogeneity.nex\".");
	}
	print(CMD '#NEXUS' . "\n\nBegin PAUP;\n\tSet ErrorBeep=No QueryBeep=No KeyBeep=No NotifyBeep=No ErrorStop=No WarnReset=No WarnTree=No WarnTSave=No WarnBlkName=No WarnRoot=No WarnRedef=No;\n\tExecute data_$partname.nex;\n\tDefaults LScores LongFmt=Yes;\n\tSet Criterion=Likelihood;\n\tGetTrees File=starttree_");
	if ($difftreeeachpart) {
		if ($partname =~ /^(.+_P)[1-3]$/i) {
			print(CMD $1);
		} else {
			print(CMD $partname);
		}
	} else {
		print(CMD 'whole');
	}
	print(CMD ".tre From=1 To=1 Unrooted=Yes;\n");
	if ($heterogeneity eq 'arbitrary') {
		if ($partname =~ /_P$/i || $partname eq 'whole' && @partname == 1 && @subpart == 3) {
			print(CMD "\tCharPartition Kakusan = 1:1-.\\3, 2:2-.\\3, 3:3-.\\3;\n")
		} elsif ($partname eq 'whole' && @partname > 1) {
			my @temppartname;
			foreach my $temp (@partname) {
				if ($temp =~ /_P$/i) {
					foreach my $subpartno (1,2,3) {
						push(@temppartname, $temp . $subpartno);
						print(CMD "\tCharSet " . $temp . $subpartno . '=' . "@{$subpart{$temp . $subpartno}}" . ';' . "\n");
					}
				} else {
					push(@temppartname, $temp);
					print(CMD "\tCharSet $temp = @{$part2site{$temp}};\n");
				}
			}
			print(CMD "\tCharPartition Kakusan = ");
			for (my $i = 0; $i < @temppartname; $i ++) {
				print(CMD $i + 1 . ":$temppartname[$i]");
				if ($i + 1 == @temppartname) {
					print(CMD ";\n");
					last;
				} else {
					print(CMD ', ');
				}
			}
		}
	}
	print(CMD "\t" . &makePAUPLSet($model, $heterogeneity) . "\n");
	print(CMD "\tLScores 1 / ScoreFile=Scores/" . $partname . '_' . $model . '_' . $heterogeneity . ".txt Replace=Yes;\n");
	print(CMD "\tQuit WarnTSave=No;\n");
	print(CMD 'End;' . "\n");
	close(CMD);
	if (system("$paupcmd " . $partname . '_' . $model . '_' . $heterogeneity . '.nex > Logs/' . $partname . '_' . $model . '_' . $heterogeneity . '.txt')) {
		&errorMessage("Cannot run $paupcmd. Please check $paupcmd.");
	}
	unless (rename($partname . '_' . $model . '_' . $heterogeneity . '.nex', 'Logs/' . $partname . '_' . $model . '_' . $heterogeneity . '.nex')) {
		&errorMessage("Cannot move \"$maindir/" . $partname . '_' . $model . '_' . $heterogeneity . ".nex\".");
	}
}

sub calculateCriteria {
	print("Calculating AIC, AICc, and BIC...\n");
	unless (chdir($maindir)) {
		&errorMessage('Cannot change working directory.');
	}
	foreach my $partname (@allpart) {
		foreach my $model (keys(%{$models{$partname}})) {
			foreach my $heterogeneity (keys(%{$heterogeneities{$partname}})) {
				if (open(LNL, '< Scores/' . $partname . '_' . $model . '_' . $heterogeneity . '.txt')) {
					while (<LNL>) {
						if (/^1\t(\d+\.\d+)/) {
							$LnL{$partname}{$model . '_' . $heterogeneity} = $1;
							last;
						}
					}
					close(LNL);
					$criteria{'AIC'}{$partname}{$model . '_' . $heterogeneity} = &calcAIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity});
					if ($samplesize{$partname}[0] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc1'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[0]);
					}
					if ($samplesize{$partname}[1] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc2'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[1]);
					}
					if ($samplesize{$partname}[2] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc3'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[2]);
					}
					if ($samplesize{$partname}[3] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc4'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[3]);
					}
					if ($samplesize{$partname}[4] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc5'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[4]);
					}
					if ($samplesize{$partname}[5] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc6'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[5]);
					}
					$criteria{'BIC1'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[0]);
					$criteria{'BIC2'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[1]);
					$criteria{'BIC3'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[2]);
					$criteria{'BIC4'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[3]);
					$criteria{'BIC5'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[4]);
					$criteria{'BIC6'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[5]);
				}
			}
		}
	}
	&calcweight();
	unless (chdir('..')) {
		&errorMessage('Cannot change working directory.');
	}
	print("done.\n\n");
}

sub calcAIC {
	my $AIC = 2 * $_[0] + 2 * $_[1];
	return($AIC);
}

sub calcAICc {
	my $AICc;
	$AICc = 2 * $_[0] + 2 * $_[1] * $_[2] / ($_[2] - $_[1] - 1);
	return($AICc);
}

sub calcBIC {
	my $BIC = 2 * $_[0] + $_[1] * log($_[2]);
	return($BIC);
}

sub calcweight {
	foreach my $partname (@allpart) {
		foreach my $criterion (keys(%criteria)) {
			my $num = 0;
			my $minvalue;
			my $sum = 0;
			foreach my $model (sort({$criteria{$criterion}{$partname}{$a} <=> $criteria{$criterion}{$partname}{$b}} keys(%{$criteria{$criterion}{$partname}}))) {
				if ($num == 0) {
					$minvalue = $criteria{$criterion}{$partname}{$model};
					$sum = 1;
				} else {
					$sum += exp(($criteria{$criterion}{$partname}{$model} - $minvalue) / (-2));
				}
				$num ++;
			}
			foreach my $model (keys(%{$criteria{$criterion}{$partname}})) {
				$weight{$criterion}{$partname}{$model} = exp(($criteria{$criterion}{$partname}{$model} - $minvalue) / (-2)) / $sum;
			}
		}
	}
}

sub outputFiles {
	print("Outputting results...\n");
	unless (chdir($maindir)) {
		&errorMessage('Cannot change working directory.');
	}
	# output results
	&outputResults();
	# output for MrBayes
	if ($output{'mrbayes'} || $output{'all'}) {
		&outforMrBayes();
	}
	# output for PAUP*
	if ($output{'paup'} || $output{'all'}) {
		&outforPAUP();
	}
	unless (chdir('..')) {
		&errorMessage('Cannot change working directory.');
	}
	print("done.\n\n");
}

sub outputResults {
	unless (mkdir('Results')) {
		&errorMessage("Cannot make \"$maindir/Results\" directory.");
	}
	foreach my $partname (@allpart) {
		foreach my $criterion (keys(%criteria)) {
			if ($criteria{$criterion}{$partname}{'JC69_homogeneous'}) {
				unless (open(OUT, "> Results/$partname" . "_$criterion.txt")) {
					&errorMessage("Cannot make \"$maindir/Resilts/$partname" . '_' . "$criterion.txt\".");
				}
				print(OUT 'model                ' . "\t");
				printf(OUT "% 19s \t", $criterion);
				print(OUT 'weight ' . "\t");
				print(OUT '               -LnL ' . "\t" . 'nparam');
				if ($criterion =~ /(\d)$/) {
					print(OUT "\t" . 'sample size' );
				}
				print(OUT "\n");
				foreach my $model (sort({$criteria{$criterion}{$partname}{$a} <=> $criteria{$criterion}{$partname}{$b}} keys(%{$criteria{$criterion}{$partname}}))) {
					printf(OUT "%-21s\t", $model);
					printf(OUT "% 20.12e\t", $criteria{$criterion}{$partname}{$model});
					printf(OUT "%.5f\t", $weight{$criterion}{$partname}{$model});
					printf(OUT "% 20.12e\t", $LnL{$partname}{$model});
					printf(OUT "% 6d", $nparam{$partname}{$model});
					if ($criterion =~ /(\d)$/) {
						printf(OUT "\t% 11d", $samplesize{$partname}[$1 - 1]);
					}
					print(OUT "\n");
				}
				close(OUT);
			}
		}
	}
}

sub outforPAUP {
	unless (mkdir('PAUP')) {
		&errorMessage("Cannot make \"$maindir/PAUP\" directory.");
	}
	my @temppartname;
	if (@partname > 1) {
		@temppartname = @partname;
	}
	foreach my $criterion (keys(%criteria)) {
		foreach my $partname ('whole', @temppartname) {
			if ($criteria{$criterion}{$partname}{'JC69_homogeneous'}) {
				&makePAUPNexus($criterion, $partname);
			}
		}
	}
}

sub makePAUPNexus {
	my ($criterion, $partname) = @_;
	my %modelPAUP;
	foreach my $model (@modelforpaup) {
		foreach my $heterogeneity (@heterogeneityforpaup) {
			$modelPAUP{$model . '_' . $heterogeneity} = 1;
		}
	}
	unless (open(OUT, '> PAUP/' . $partname . '_' . $criterion . ".nex")) {
		&errorMessage("Cannot make \"$maindir/PAUP/$partname" . '_' . "$criterion.nex\".");
	}
	print(OUT "#NEXUS\n\n");
	print(OUT "Begin PAUP;\n");
	print(OUT "\tExecute ../data_$partname.nex;\n");
	if ($partname =~ /_P$/i || $partname eq 'whole' && @partname == 1 && @subpart == 3) {
		print(OUT "\tCharPartition Kakusan = 1:1-.\\3, 2:2-.\\3, 3:3-.\\3;\n")
	} elsif ($partname eq 'whole' && @partname > 1) {
		my @temppartname;
		foreach my $temp (@partname) {
			if ($temp =~ /_P$/i) {
				foreach my $subpartno (1,2,3) {
					push(@temppartname, $temp . $subpartno);
					print(OUT "\tCharSet " . $temp . $subpartno . '=' . "@{$subpart{$temp . $subpartno}}" . ';' . "\n");
				}
			} else {
				push(@temppartname, $temp);
				print(OUT "\tCharSet $temp = @{$part2site{$temp}};\n");
			}
		}
		print(OUT "\tCharPartition Kakusan = ");
		for (my $i = 0; $i < @temppartname; $i ++) {
			print(OUT $i + 1 . ":$temppartname[$i]");
			if ($i + 1 == @temppartname) {
				print(OUT ";\n");
				last;
			} else {
				print(OUT ', ');
			}
		}
	}
	print(OUT "\tSet Criterion=Likelihood;\n");
	my $num = 0;
	foreach my $model (sort({$criteria{$criterion}{$partname}{$a} <=> $criteria{$criterion}{$partname}{$b}} keys(%{$criteria{$criterion}{$partname}}))) {
		print(OUT '[ ');
		printf(OUT "%-21s ", $model);
		printf(OUT "% 12.5e", $criteria{$criterion}{$partname}{$model});
		print(OUT ' ]  ');
		if ($modelPAUP{$model}) {
			if ($num == 0) {
				print(OUT '  ');
			} else {
				print(OUT '[ ');
			}
			print(OUT &makePAUPLSet(split(/_/, $model)));
			if ($num != 0) {
				print(OUT ' ]');
			}
			$num ++;
		}
		print(OUT "\n");
	}
	print(OUT "End;\n");
	close(OUT);
}

sub makePAUPLSet {
	my ($model, $heterogeneity) = @_;
	my $lset1;
	my $lset2;
	if ($model eq 'JC69') {
		$lset1 = 'Nst=1 BaseFreq=Equal';
	} elsif ($model eq 'F81') {
		$lset1 = 'Nst=1 BaseFreq=' . $basefreq;
	} elsif ($model eq 'K80') {
		$lset1 = 'Nst=2 BaseFreq=Equal TRatio=Estimate';
	} elsif ($model eq 'HKY85') {
		$lset1 = 'Nst=2 BaseFreq=' . $basefreq . ' TRatio=Estimate';
	} elsif ($model eq 'TN93ef') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b a a e a)';
	} elsif ($model eq 'TN93') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b a a e a)';
	} elsif ($model eq 'K81uf') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b c c b a)';
	} elsif ($model eq 'K81') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b c c b a)';
	} elsif ($model eq 'TIMef') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b c c e a)';
	} elsif ($model eq 'TIM') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b c c e a)';
	} elsif ($model eq 'TVMef') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b c d b e)';
	} elsif ($model eq 'TVM') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b c d b e)';
	} elsif ($model eq 'SYM') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b c d e f)';
	} elsif ($model eq 'GTR') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b c d e f)';
	}
	if ($heterogeneity eq 'homogeneous') {
		$lset2 = ' Rates=Equal PInvar=0';
	} elsif ($heterogeneity eq 'gamma') {
		$lset2 = ' Rates=Gamma Shape=Estimate ' . "NCat=$ngammacat";
	} elsif ($heterogeneity eq 'invariant') {
		$lset2 = ' Rates=Equal PInvar=Estimate';
	} elsif ($heterogeneity eq 'gammainvariant') {
		$lset2 = ' Rates=Gamma Shape=Estimate PInvar=Estimate ' . "NCat=$ngammacat";
	} elsif ($heterogeneity eq 'arbitrary') {
		$lset2 = ' Rates=SiteSpec SiteRates=Partition:Kakusan';
	}
	my $cmd = "LSet $lset1$lset2;";
	return($cmd);
}

sub outforMrBayes {
	unless (mkdir('MrBayes')) {
		&errorMessage("Cannot make \"$maindir/MrBayes\" directory.");
	}
	my @temppartname;
	if (@partname > 1) {
		@temppartname = @partname;
	}
	foreach my $criterion (keys(%criteria)) {
		foreach my $partname ('whole', @temppartname) {
			if ($criteria{$criterion}{$partname}{'JC69_homogeneous'}) {
				&makeMrBayesNexus($criterion, $partname);
			}
		}
	}
}

sub makeMrBayesNexus {
	my ($criterion, $partname) = @_;
	my $outfile;
	my $partcriterion = 1;
	my $subpartcriterion = 1;
	if ($partname eq 'whole') {
		foreach my $temppartname (@partname) {
			unless ($criteria{$criterion}{$temppartname}{'JC69_homogeneous'}) {
				$partcriterion = 0;
			}
		}
		foreach my $temppartname (@subpart) {
			unless ($criteria{$criterion}{$temppartname}{'JC69_homogeneous'}) {
				$subpartcriterion = 0;
			}
		}
	} elsif ($partname =~ /_P$/i) {
		foreach my $subpartno (1,2,3) {
			unless ($criteria{$criterion}{$partname . $subpartno}{'JC69_homogeneous'}) {
				$subpartcriterion = 0;
			}
		}
	}
	# single model
	if ($partname eq 'whole' && @partname == 1 && @subpart > 0 || $partname =~ /_P$/i) {
		$outfile = $partname . '_' . $criterion . '_codonshared.nex';
	} else {
		$outfile = $partname . '_' . $criterion . '_single.nex';
	}
	if ($outfile) {
		open(OUT, "> MrBayes/$outfile");
		print(OUT "#NEXUS\n\n");
		print(OUT "Begin MrBayes;\n\tExecute ../data_$partname.nex;\n");
		my @temppartname;
		if ($partname eq 'whole') {
			foreach my $temppartname (@partname) {
				if ($temppartname =~ /_P$/i) {
					foreach my $subpartno (1,2,3) {
						push(@temppartname, $temppartname . $subpartno);
						print(OUT "\tCharSet $temppartname$subpartno=@{$subpart{$temppartname . $subpartno}};\n");
					}
				} else {
					push(@temppartname, $temppartname);
					print(OUT "\tCharSet $temppartname=@{$part2site{$temppartname}};\n");
				}
			}
		} elsif ($partname =~ /_P$/i) {
			foreach my $subpartno (1,2,3) {
				push(@temppartname, $partname . $subpartno);
				print(OUT "\tCharSet $partname$subpartno=$subpartno-.\\3;\n");
			}
		}
		if (@temppartname > 0 && $partname eq 'whole' || @temppartname > 0 && $partname =~ /_P$/i) {
			print(OUT "\tPartition Kakusan=" . scalar(@temppartname) . ':' . join(',', @temppartname) . ';' . "\n");
			print(OUT "\tSet Partition=Kakusan;\n");
		}
		print(OUT &makeMrBayesCommand($criterion, $partname));
		print(OUT "\tMCMCP Filename=$outfile;\n");
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
	# proportional model (codon position singlized)
	if ($partname eq 'whole' && @partname > 1 && $partcriterion == 1) {
		if (@subpart > 0) {
			$outfile = $partname . '_' . $criterion . '_proportional_codonshared.nex';
		} else {
			$outfile = $partname . '_' . $criterion . '_proportional.nex';
		}
		open(OUT, "> MrBayes/$outfile");
		print(OUT "#NEXUS\n\n");
		print(OUT "Begin MrBayes;\n\tExecute ../data_$partname.nex;\n");
		my $partno = 0;
		my $num = 0;
		my @applyto;
		my @temppartname;
		foreach my $temppartname (@partname) {
			if ($outfile =~ /codonshared.nex$/) {
				my %modelMrBayes;
				foreach my $tempmodel (@modelformrbayes) {
					foreach my $tempheterogeneity (@heterogeneityformrbayes) {
						$modelMrBayes{$tempmodel . '_' . $tempheterogeneity} = 1;
					}
				}
				foreach my $tempmodel (sort({$criteria{$criterion}{$temppartname}{$a} <=> $criteria{$criterion}{$temppartname}{$b}} keys(%{$criteria{$criterion}{$temppartname}}))) {
					if ($modelMrBayes{$tempmodel}) {
						unless ($tempmodel =~ /arbitrary$/) {
							$notarbitrary{$temppartname} = 1;
						}
						last;
					}
				}
			}
			unless ($notarbitrary{$temppartname}) {
				foreach my $subpartno (1,2,3) {
					$partno ++;
					push(@{$applyto[$num]}, $partno);
					push(@temppartname, $temppartname . $subpartno);
					print(OUT "\tCharSet $temppartname$subpartno=@{$subpart{$temppartname . $subpartno}};\n");
				}
				$num ++;
			} else {
				$partno ++;
				push(@temppartname, $temppartname);
				print(OUT "\tCharSet $temppartname=@{$part2site{$temppartname}};\n");
			}
		}
		print(OUT "\tPartition Kakusan=" . scalar(@temppartname) . ':' . join(',', @temppartname) . ';' . "\n");
		print(OUT "\tSet Partition=Kakusan;\n");
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all);\n");
		for (my $i = 0; $i < @applyto; $i ++) {
			print(OUT "\tLink TRatio=(" . join(',', @{$applyto[$i]}) . ") RevMat=(" . join(',', @{$applyto[$i]}) . ") StateFreq=(" . join(',', @{$applyto[$i]}) . ") Shape=(" . join(',', @{$applyto[$i]}) . ") PInvar=(" . join(',', @{$applyto[$i]}) . ");\n");
		}
		print(OUT "\tPrSet ApplyTo=(all) RatePr=Variable;\n");
		print(OUT &makeMrBayesCommand($criterion, @partname));
		print(OUT "\tMCMCP Filename=$outfile;\n");
		print(OUT "End;\n");
		close(OUT);
		undef(%notarbitrary);
		undef($outfile);
	}
	# proportional model (codon position proportional)
	if ($partname eq 'whole' && @partname > 1 && @subpart > 0 && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_proportional_codonproportional.nex';
	} elsif ($partname eq 'whole' && @subpart > 0 && $subpartcriterion == 1 || $partname =~ /_P$/i && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_codonproportional.nex';
	}
	if ($outfile) {
		open(OUT, "> MrBayes/$outfile");
		print(OUT "#NEXUS\n\n");
		print(OUT "Begin MrBayes;\n\tExecute ../data_$partname.nex;\n");
		my @temppartname;
		if ($partname eq 'whole') {
			foreach my $temppartname (@partname) {
				if ($temppartname =~ /_P$/i) {
					foreach my $subpartno (1,2,3) {
						push(@temppartname, $temppartname . $subpartno);
						print(OUT "\tCharSet $temppartname$subpartno=@{$subpart{$temppartname . $subpartno}};\n");
					}
				} else {
					push(@temppartname, $temppartname);
					print(OUT "\tCharSet $temppartname=@{$part2site{$temppartname}};\n");
				}
			}
		} else {
			foreach my $subpartno (1,2,3) {
				push(@temppartname, $partname . $subpartno);
				print(OUT "\tCharSet $partname$subpartno=$subpartno-.\\3;\n");
			}
		}
		print(OUT "\tPartition Kakusan=" . scalar(@temppartname) . ':' . join(',', @temppartname) . ';' . "\n");
		print(OUT "\tSet Partition=Kakusan;\n");
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all);\n");
		print(OUT "\tPrSet ApplyTo=(all) RatePr=Variable;\n");
		print(OUT &makeMrBayesCommand($criterion, @temppartname));
		print(OUT "\tMCMCP Filename=$outfile;\n");
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
	# proportional model (codon position separate)
#	if ($partname eq 'whole' && @partname > 1 && @subpart > 0 && @subpart != 3 * @partname && $subpartcriterion == 1) {
#		$outfile = $partname . '_' . $criterion . '_proportional_codonseparate.nex';
#		open(OUT, "> MrBayes/$outfile");
#		print(OUT "#NEXUS\n\n");
#		print(OUT "Begin MrBayes;\n\tExecute ../data_$partname.nex;\n");
#		my $partno = 0;
#		my @applyto;
#		my @temppartname;
#		foreach my $temppartname (@partname) {
#			if ($temppartname =~ /_P$/i) {
#				foreach my $subpartno (1,2,3) {
#					$partno ++;
#					push(@temppartname, $temppartname . $subpartno);
#					print(OUT "\tCharSet $temppartname$subpartno=@{$subpart{$temppartname . $subpartno}};\n");
#				}
#			} else {
#				$partno ++;
#				push(@applyto, $partno);
#				push(@temppartname, $temppartname);
#				print(OUT "\tCharSet $temppartname=@{$part2site{$temppartname}};\n");
#			}
#		}
#		print(OUT "\tPartition Kakusan=" . scalar(@temppartname) . ':' . join(',', @temppartname) . ';' . "\n");
#		print(OUT "\tSet Partition=Kakusan;\n");
#		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) BrLens=(all);\n");
#		print(OUT "\tLink BrLens=(" . join(',', @applyto) . ");\n");
#		print(OUT "\tPrSet ApplyTo=(" . join(',', @applyto) . ") RatePr=Variable;\n");
#		print(OUT &makeMrBayesCommand($criterion, @temppartname));
#		print(OUT "\tMCMCP Filename=$outfile;\n");
#		print(OUT "End;\n");
#		close(OUT);
#		undef($outfile);
#	}
	# separate model (codon position singlized)
	if ($partname eq 'whole' && @partname > 1 && $partcriterion == 1) {
		if (@subpart > 0) {
			$outfile = $partname . '_' . $criterion . '_separate_codonshared.nex';
		} else {
			$outfile = $partname . '_' . $criterion . '_separate.nex';
		}
		open(OUT, "> MrBayes/$outfile");
		print(OUT "#NEXUS\n\n");
		print(OUT "Begin MrBayes;\n\tExecute ../data_$partname.nex;\n");
		my $partno = 0;
		my $num = 0;
		my @applyto;
		my @temppartname;
		foreach my $temppartname (@partname) {
			if ($temppartname =~ /_P$/i) {
				foreach my $subpartno (1,2,3) {
					$partno ++;
					push(@{$applyto[$num]}, $partno);
					push(@temppartname, $temppartname . $subpartno);
					print(OUT "\tCharSet $temppartname$subpartno=@{$subpart{$temppartname . $subpartno}};\n");
				}
				$num ++;
			} else {
				$partno ++;
				push(@temppartname, $temppartname);
				print(OUT "\tCharSet $temppartname=@{$part2site{$temppartname}};\n");
			}
		}
		print(OUT "\tPartition Kakusan=" . scalar(@temppartname) . ':' . join(',', @temppartname) . ';' . "\n");
		print(OUT "\tSet Partition=Kakusan;\n");
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) BrLens=(all);\n");
		for (my $i = 0; $i < @applyto; $i ++) {
			print(OUT "\tLink TRatio=(" . join(',', @{$applyto[$i]}) . ") RevMat=(" . join(',', @{$applyto[$i]}) . ") StateFreq=(" . join(',', @{$applyto[$i]}) . ") Shape=(" . join(',', @{$applyto[$i]}) . ") PInvar=(" . join(',', @{$applyto[$i]}) . ") BrLens=(" . join(',', @{$applyto[$i]}) . ");\n");
		}
		print(OUT &makeMrBayesCommand($criterion, @partname));
		print(OUT "\tMCMCP Filename=$outfile;\n");
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
	# separate model (codon position proportional)
	if ($partname eq 'whole' && @partname > 1 && @subpart > 0 && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_separate_codonproportional.nex';
		open(OUT, "> MrBayes/$outfile");
		print(OUT "#NEXUS\n\n");
		print(OUT "Begin MrBayes;\n\tExecute ../data_$partname.nex;\n");
		my $partno = 0;
		my $num = 0;
		my @applyto;
		my @temppartname;
		foreach my $temppartname (@partname) {
			if ($temppartname =~ /_P$/i) {
				foreach my $subpartno (1,2,3) {
					$partno ++;
					push(@{$applyto[$num]}, $partno);
					push(@temppartname, $temppartname . $subpartno);
					print(OUT "\tCharSet $temppartname$subpartno=@{$subpart{$temppartname . $subpartno}};\n");
				}
				$num ++;
			} else {
				$partno ++;
				push(@temppartname, $temppartname);
				print(OUT "\tCharSet $temppartname=@{$part2site{$temppartname}};\n");
			}
		}
		print(OUT "\tPartition Kakusan=" . scalar(@temppartname) . ':' . join(',', @temppartname) . ';' . "\n");
		print(OUT "\tSet Partition=Kakusan;\n");
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) BrLens=(all);\n");
		for (my $i = 0; $i < @applyto; $i ++) {
			print(OUT "\tLink BrLens=(" . join(',', @{$applyto[$i]}) . ");\n");
			print(OUT "\tPrSet ApplyTo=(" . join(',', @{$applyto[$i]}) . ") RatePr=Variable;\n");
		}
		print(OUT &makeMrBayesCommand($criterion, @temppartname));
		print(OUT "\tMCMCP Filename=$outfile;\n");
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
	# separate model (codon position separate)
	if ($partname eq 'whole' && @partname > 1 && @subpart > 0 && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_separate_codonseparate.nex';
	} elsif ($partname eq 'whole' && @subpart > 0 && $subpartcriterion == 1 || $partname =~ /_P$/i && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_codonseparate.nex';
	}
	if ($outfile) {
		open(OUT, "> MrBayes/$outfile");
		print(OUT "#NEXUS\n\n");
		print(OUT "Begin MrBayes;\n\tExecute ../data_$partname.nex;\n");
		my @temppartname;
		if ($partname eq 'whole') {
			foreach my $temppartname (@partname) {
				if ($temppartname =~ /_P$/i) {
					foreach my $subpartno (1,2,3) {
						push(@temppartname, $temppartname . $subpartno);
						print(OUT "\tCharSet $temppartname$subpartno=@{$subpart{$temppartname . $subpartno}};\n");
					}
				} else {
					push(@temppartname, $temppartname);
					print(OUT "\tCharSet $temppartname=@{$part2site{$temppartname}};\n");
				}
			}
		} else {
			foreach my $subpartno (1,2,3) {
				push(@temppartname, $partname . $subpartno);
				print(OUT "\tCharSet $partname$subpartno=$subpartno-.\\3;\n");
			}
		}
		print(OUT "\tPartition Kakusan=" . scalar(@temppartname) . ':' . join(',', @temppartname) . ';' . "\n");
		print(OUT "\tSet Partition=Kakusan;\n");
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) BrLens=(all);\n");
		print(OUT &makeMrBayesCommand($criterion, @temppartname));
		print(OUT "\tMCMCP Filename=$outfile;\n");
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
}

sub makeMrBayesCommand {
	my %modelMrBayes;
	foreach my $model (@modelformrbayes) {
		foreach my $heterogeneity (@heterogeneityformrbayes) {
			$modelMrBayes{$model . '_' . $heterogeneity} = 1;
		}
	}
	my $partno = 1;
	my $criterion = shift(@_);
	my $cmd;
	foreach my $partname (@_) {
		$cmd .= '[' . $partname . ']' . "\n";
		my $num = 0;
		foreach my $model (sort({$criteria{$criterion}{$partname}{$a} <=> $criteria{$criterion}{$partname}{$b}} keys(%{$criteria{$criterion}{$partname}}))) {
			$cmd .= '[ ' . sprintf("%-21s ", $model);
			$cmd .= sprintf("% 12.5e", $criteria{$criterion}{$partname}{$model});
			$cmd .= ' ]  ';
			if ($modelMrBayes{$model}) {
				if ($num == 0) {
					$cmd .= '  ';
				} else {
					$cmd .= '[ ';
				}
				$cmd .= &makeMrBayesLSet($model, $partno, $partname);
				if ($notarbitrary{$partname}) {
					last;
				}
				if ($num != 0) {
					$cmd .=' ]';
				}
				$num ++;
			}
			$cmd .= "\n";
		}
		if ($partname =~ /_P$/i && ! $notarbitrary{$partname}) {
			$partno += 3;
		} else {
			$partno ++;
		}
	}
	return($cmd);
}

sub makeMrBayesLSet {
	my ($model, $heterogeneity) = split(/_/, $_[0]);
	my $partno = $_[1];
	my $partname = $_[2];
	my $mblset1;
	my $mblset2;
	my $mbprset1;
	my $mbprset2;
	if ($model eq 'JC69') {
		$mblset1 = 'Nst=1';
		$mbprset1 = 'StateFreqPr=Fixed(Equal)';
	} elsif ($model eq 'F81') {
		$mblset1 = 'Nst=1';
		$mbprset1 = 'StateFreqPr=Dirichlet(1,1,1,1)';
	} elsif ($model eq 'K80') {
		$mblset1 = 'Nst=2';
		$mbprset1 = 'StateFreqPr=Fixed(Equal)';
	} elsif ($model eq 'HKY85') {
		$mblset1 = 'Nst=2';
		$mbprset1 = 'StateFreqPr=Dirichlet(1,1,1,1)';
	} elsif ($model eq 'SYM') {
		$mblset1 = 'Nst=6';
		$mbprset1 = 'StateFreqPr=Fixed(Equal)';
	} elsif ($model eq 'GTR') {
		$mblset1 = 'Nst=6';
		$mbprset1 = 'StateFreqPr=Dirichlet(1,1,1,1)';
	}
	if ($heterogeneity eq 'homogeneous') {
		$mblset2 = 'Rates=Equal';
	} elsif ($heterogeneity eq 'gamma') {
		$mblset2 = 'Rates=Gamma NGammaCat=' . $ngammacat;
	} elsif ($heterogeneity eq 'invariant') {
		$mblset2 = 'Rates=PropInv';
	} elsif ($heterogeneity eq 'gammainvariant') {
		$mblset2 = 'Rates=InvGamma NGammaCat=' . $ngammacat;
	} elsif ($heterogeneity eq 'arbitrary') {
		$mblset2 = 'Rates=Equal';
		$mbprset2 = ' RatePr=Variable';
	}
	my @applyto;
	my $repeat = 1;
	if ($partname =~ /_P$/i && ! $notarbitrary{$partname} || $partname eq 'whole' && @partname == 1 && @subpart == 3) {
		$repeat = 3;
	} elsif ($partname eq 'whole' && @partname > 1) {
		$repeat = @partname - (@subpart / 3) + @subpart;
	}
	for (my $i = 0; $i < $repeat; $i ++) {
		push(@applyto, $partno + $i);
	}
	my $cmd = "LSet ApplyTo=(" . join(',', @applyto) . ") NucModel=4by4 $mblset1 $mblset2; PrSet ApplyTo=(" . join(',', @applyto) . ") $mbprset1$mbprset2;";
	return($cmd);
}

sub errorMessage {
	print("ERROR!: @_\n");
	print("If you want to read help message, run this script with \"--help\" option.\n");
	exit(1);
}

sub helpMessage {
	print << "_END";
Usage
=====

perl kakusan.pl [-hcsdfpgaton] inputfile

Command line options
====================

-h, --help
  Display this message and exit.

-c, --chisq=Enable|Disable
  Specify whether you want to test base frequencies homogeneity across
taxa or not. (default: Enable)

-s, --siglevel=DECIMAL
  Specify significant level of chi-square test. (default: 0.05)

-w, --calcwhole=Enable|Disable
  Specify whether you want to consider single models on whole
sequences or not. (default: Enable)

-d, --dividecodon=Enable|Disable
  Specify whether you want to select a model on each codon position of
protein coding region or not. (default: Enable)

-f, --basefreq=Empirical|Estimate
  Specify base frequencies optimization approach. (default: Estimate)

-p, --partition=PARTITIONS
  Specify data partitions. (default: none)

-g, --ngammacat=INTEGER
  Specify the number of gamma rate categories. (default: 4)

-a, --algorithm=NJJC|NJLogDet|SARandom|SAClosest|SAFurthest
  Specify tree generating algorithm for parameter optimization.
(default: NJJC)

-t, --tree=Common|Different
  If you specify 'Common', the parameters are optimized at a common
tree on each partition. If you specify 'Different', the parameters are
optimized at different tree on each partition. (default: Common)

-o, --output=MrBayes|PAUP|All
  Specify output options. (default: All)

-n, --numthreads=INTEGER
  Specify the number of threads. (default: 1)

Examples of data partition specification
=========================================

part1:1-200,part2:201-500,part3:501-550
part1_P:1-300
  \"_P\" means \"protein coding region\".
_END
	exit(0);
}
